#ifndef _RaveInterface_VertexFactory_H_
#define _RaveInterface_VertexFactory_H_

#include <rave/RaveTrack.h>
#include <rave/RaveVertex.h>
#include <rave/RaveMagneticField.h>
#include <rave/RavePropagator.h>
#include <rave/RaveEllipsoid3D.h>
#include <rave/RaveVector3D.h>
#include <rave/VacuumPropagator.h>
#include <rave/ConstantMagneticField.h>
#include <vector>
#include <map>
#include <string>

class RaveReconstructor;

namespace rave
{
/** 
 *  The class that is central to using the rave library.
 *  It builds the RaveVertices from a set of RaveTracks.
 *  \paramname method is a string that describes which
 *  algorithms should be employed and which parameter
 *  settings should be chosen.
 *  The default is chosen such that something
 *  "sensible" is done in most use cases.
 */

class VertexFactory 
{
  public:
    VertexFactory ( const rave::MagneticField & m = ConstantMagneticField(),
              const rave::Propagator & p = VacuumPropagator(),
              const std::string & method = "default",
              int verbosity=1 );
    
    VertexFactory ( const rave::MagneticField &,
              const rave::Propagator &,
              const rave::Ellipsoid3D & beamspot,
              const std::string & method = "default",
              int verbosity=1 );

    VertexFactory ( const VertexFactory & o );
    ~VertexFactory();
    VertexFactory operator= ( const VertexFactory & o );

    /** reconstruct vertices, using the default reconstruction method
     * \paramname use_beamspot defines, if the beamspot information
     * is taken into account.
     */
    std::vector < Vertex > create ( const std::vector < Track > &, bool use_beamspot=false ) const;

    /** reconstruct vertices, using "method".
     * \paramname use_beamspot defines, if the beamspot information
     * is taken into account.
     */
    std::vector < Vertex > create ( const std::vector < Track > &, 
                                    const std::string & method, bool use_beamspot=false ) const;

    /// returns the propagator that is used
    const rave::Propagator & getPropagator() const;

    /// returns the magnetic field
    const rave::MagneticField & getMagneticField() const;

    /// returns the beamspot constraint that is used
    const rave::Ellipsoid3D & getBeamspot() const;

    /// true, if the factory knows of a beamspot constraint
    bool hasBeamspot() const;

    const int verbosity() const;

    /// returns the default vertex reconstruction method
    const std::string & method() const;

    /// set the default vertex reconstruction method
    void setDefaultMethod ( const std::string & );
    
    const RaveReconstructor & reconstructor() const;

    /**
     * a virtual init, for subclasses.
     */
    virtual void init() {};

  private:
    void setup();
    std::vector < Vertex > fit ( const std::vector < Track > &, const RaveReconstructor &,
                                 bool use_bs ) const;
    void createRecoBeamSpot();

  private:
    RaveReconstructor * theRector;
    mutable std::map < std::string, RaveReconstructor * > theRectors;

    std::string theMethod;
    rave::MagneticField * theField;
    rave::Propagator * theProp;
    rave::Ellipsoid3D theBeamspot;
    void * theRecoBeamSpot; //< the reco::BeamSpot, cast to void *
    int theVerbosity;
};

}

#endif
